"""
@file Unpacker.py
@author Mickael Lehoux <https://github.com/MkLHX>
@brief Class to allow raspberry I2C Master to deal with ESP32
using ESP32 Slave I2C library
@date 2020-09-25

The ESP32 Slave I2C library
use packing and upacking classes to format data
On python side we need to DECODE data from i2c

format:
    [0]: start byte (0x02)
    [1]: packet length
    [2]: data[0]
    [3]: data[1]
    ...
    [n+1]: data[n-1]
    [n+2]: CRC8 of packet length and data
    [n+3]: end byte (0x04)
based on:
https://github.com/gutierrezps/ESP32_I2C_Slave/blob/master/src/WireUnpacker.h
https://github.com/gutierrezps/ESP32_I2C_Slave/blob/master/src/WireUnpacker.cpp
"""
from Raspberry_Pi_Master_for_ESP32_I2C_SLAVE.crc8 import Crc8


class Unpacker:
    error_codes = {
        "INVALID_CRC": 1,
        "INVALID_LENGTH": 2,
        "INVALID_START": 5,
        "INVALID_END": 6,
    }
    error_decodes = {
        1: "INVALID_CRC",
        2: "INVALID_LENGTH",
        3: "INVALID_START",
        4: "INVALID_END",
    }

    def __init__(self):
        self._debug = False
        self.reset()

    @property
    def debug(self):
        return self._debug

    @debug.setter
    def debug(self, d):
        self._debug = d

    def __enter__(self):
        """Context manager enter function."""
        # Just return this object so it can be used in a with statement, like
        # with Unpacker() as unpacker:
        #     # do stuff!
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Context manager exit function, ensures resources are cleaned up."""
        return False  # Don't suppress exceptions.

    def __del__(self):
        """Clean up any resources instance."""
        return False

    def get_last_error(self):
        """
        @brief get the last error code and message
        @return list [error_code, error_text]
        """
        return self._last_error, self.error_decodes[self._last_error]

    def read(self):
        """
        @brief return i2c values from slave
        @return list buffer
        """
        return self._buffer

    def write(self, data: list):
        """
        @brief get the i2c data from slave
        check if start, end and crc8 bytes are correct
        remove start, length, crc8 and end bytes
        """
        if self._debug:
            print("Data to unpack: ", data)
        if data[0] != self._frame_start:
            self._last_error = self.error_codes["INVALID_START"]
            raise Exception("ERROR: Unpacker invalid start byte")
        if data[-1] != self._frame_end:
            self._last_error = self.error_codes["INVALID_END"]
            raise Exception("ERROR: Unpacker invalid end byte")

        # check if provided crc8 is good
        # ignore crc and end bytes
        payload_range = len(data) - 2
        crc8 = Crc8()
        # ignore start, length
        crc = crc8.calc(data[2:payload_range])
        if crc != data[-2]:
            self._last_error = self.error_codes["INVALID_CRC"]
            raise Exception("ERROR: Unpacker invalid crc8")

        self._buffer = data[2:-2]

    def reset(self):
        self._frame_start = 0x02
        self._frame_end = 0x04
        self._buffer = []
        self._last_error = None



# the inclusion of the tests module is not meant to offer best practices for
# testing in general, but rather to support the `find_packages` example in
# setup.py that excludes installing the "tests" package

import sys
import unittest
from Raspberry_Pi_Master_for_ESP32_I2C_SLAVE.packer import Packer


class TestPacker(unittest.TestCase):
    def test_packer(self):
        packed = None
        with Packer() as packer:
            packer.write(127)
            packer.end()
            packed = [i for i in packer.read() if i != 0]
        expected = [2, 5, 127, 185, 4]
        self.assertIsNotNone(packed)
        self.assertTrue(type(packed).__name__ == "list")
        self.assertEqual(expected, packed)


if __name__ == "__main__":
    unittest.main()



# the inclusion of the tests module is not meant to offer best practices for
# testing in general, but rather to support the `find_packages` example in
# setup.py that excludes installing the "tests" package

import sys
import unittest
from Raspberry_Pi_Master_for_ESP32_I2C_SLAVE.unpacker import Unpacker


class TestPacker(unittest.TestCase):
    def test_unpacker(self):
        value_to_unpack = [2, 6, 12, 1, 19, 4]
        unpacked = None
        with Unpacker() as unpacker:
            unpacker.write(value_to_unpack)
            unpacked = unpacker.read()
        expected = [12, 1]
        self.assertIsNotNone(unpacked)
        self.assertTrue(type(unpacked).__name__ == "list")
        self.assertEqual(expected, unpacked)


if __name__ == "__main__":
    unittest.main()



"""
@file Crc8.py
@author Mickael Lehoux <https://github.com/MkLHX>
@brief build CRC8 in python
@date 2020-09-18
based on:
https://github.com/gutierrezps/ESP32_I2C_Slave/blob/master/src/WireCrc.h
"""


class Crc8:
    def __init__(self):
        self._seed = 0
        self._debug = False

    def __enter__(self):
        """Context manager enter function."""
        # Just return this object so it can be used in a with statement, like
        # with Crc8() as crc8:
        #     # do stuff!
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Context manager exit function, ensures resources are cleaned up."""
        return False  # Don't suppress exceptions.

    def calc(self, data):
        crc = self._seed
        for _byte in data:
            extract = _byte
            for j in range(8, 0, -1):
                _sum = (crc ^ extract) & 0x01
                crc >>= 1
                if _sum:
                    crc ^= 0x8C
                extract >>= 1
        return crc



# the inclusion of the tests module is not meant to offer best practices for
# testing in general, but rather to support the `find_packages` example in
# setup.py that excludes installing the "tests" package

import sys
import unittest
from Raspberry_Pi_Master_for_ESP32_I2C_SLAVE.crc8 import Crc8


class TestPacker(unittest.TestCase):
    def test_crc8(self):
        value_to_convert = [12, 1]
        _crc8 = None
        with Crc8() as crc8:
            _crc8 = crc8.calc(value_to_convert)
        expected = 19
        self.assertIsNotNone(_crc8)
        self.assertTrue(type(_crc8).__name__ == "int")
        self.assertEqual(expected, _crc8)


if __name__ == "__main__":
    unittest.main()



"""
@file Packer.py
@author Mickael Lehoux <https://github.com/MkLHX>
@brief Class to allow raspberry I2C Master to deal with ESP32
using ESP32 Slave I2C library
@date 2020-09-18

The ESP32 Slave I2C library
use packing and upacking classes to format data
On python side we need to ENCODE data before send them through i2c

Packet format:
    [0]: start byte (0x02)
    [1]: packet length
    [2]: data[0]
    [3]: data[1]
    ...
    [n+1]: data[n-1]
    [n+2]: CRC8 of packet length and data
    [n+3]: end byte (0x04)
based on:
https://github.com/gutierrezps/ESP32_I2C_Slave/blob/master/src/WirePacker.h
https://github.com/gutierrezps/ESP32_I2C_Slave/blob/master/src/WirePacker.cpp
"""
from Raspberry_Pi_Master_for_ESP32_I2C_SLAVE.crc8 import Crc8


class Packer:
    PACKER_BUFFER_LENGTH = (
        128  # because ESP Slave I2C library wait for buffer[128] size
    )

    def __init__(self):
        self._debug = False
        self._buffer = [0] * self.PACKER_BUFFER_LENGTH
        self._index = 0
        self.reset()

    @property
    def debug(self):
        return self._debug

    @debug.setter
    def debug(self, d):
        self._debug = d

    def __enter__(self):
        """Context manager enter function."""
        # Just return this object so it can be used in a with statement, like
        # with Packer() as packer:
        #     # do stuff!
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Context manager exit function, ensures resources are cleaned up."""
        self.reset()
        return False  # Don't suppress exceptions.

    def __del__(self):
        """Clean up any resources instance."""
        return False

    def read(self):
        """
        Read the next available packet byte. At each call,
        the value returned by available() will be decremented.
        @return int -1 if no bytes to read / byte value
        """
        if not self._is_written:
            raise Exception(
                "You need to finish process by using packer.end() method before read buffer"
            )
        return self._buffer

    def write(self, data: int):
        """
        @brief write data in prepared buffer
        @param int data
        """
        if self._debug:
            print("Data to unpack: ", data)
        if self._is_written:  # allow write after .end()
            self._is_written = False
        self._buffer[self._index] = data
        self._index += 1

    def end(self):
        """
        @brief Closes the packet. After that, use avaiable() and read()
        to get the packet bytes.
        """
        self._index += 1  # add field for CRC byte
        self._buffer[self._index] = self._frame_end
        self._index += 1
        self._total_length = self._index
        self._buffer[1] = self._total_length

        # ignore crc and end bytes
        payload_range = self._total_length - 2

        # ignore start and length bytes [2:payload_range]
        crc = Crc8()
        _crc8 = crc.calc(self._buffer[2:payload_range])

        self._buffer[self._index - 2] = _crc8
        self._is_written = True

    def reset(self):
        """
        @brief Reset the packing process.
        """
        self._frame_start = 0x02
        self._frame_end = 0x04
        self._buffer[0] = self._frame_start
        self._index = 2  # keep field for total lenght on index 1
        self._is_written = False



import os
import pathlib
from setuptools import setup, find_packages

# Package meta-data.
NAME = "raspberrypi-esp32-i2c"
DESCRIPTION = "use Raspberry Pi as master on ESP32 i2c \
    slave when use ESP32 i2c Slave c++ library"
URL = "https://github.com/MkLHX/Raspberry_Pi_Master_for_ESP32_I2C_SLAVE"
EMAIL = "mickael@lehoux.net"
AUTHOR = "Mickael Lehoux"
REQUIRES_PYTHON = ">=3.6.0"
VERSION = "0.0.6"

# What packages are required for this module to be executed?
REQUIRED = [
    # 'requests', 'maya', 'records',
]

# What packages are optional?
EXTRAS = {
    # 'fancy feature': ['django'],
}

here = pathlib.Path(__file__).parent.resolve()

# Get the long description from the README file
long_description = (here / "README.md").read_text(encoding="utf-8")

# Load the package's version.py module as a dictionary.
about = {}
if not VERSION:
    with open(os.path.join(here, "version.py")) as f:
        exec(f.read(), about)
else:
    about["__version__"] = VERSION

setup(
    name=NAME,
    version=about["__version__"],
    author=AUTHOR,
    author_email=EMAIL,
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type="text/markdown",
    url=URL,
    license="MIT",
    install_requires=REQUIRED,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    packages=find_packages(),
    python_requires=REQUIRES_PYTHON,
    project_urls={  # Optional
        "Source": "https://github.com/MkLHX/\
            Raspberry_Pi_Master_for_ESP32_I2C_SLAVE",
        "Bug Reports": "https://github.com/MkLHX/\
            Raspberry_Pi_Master_for_ESP32_I2C_SLAVE/issues",
    },
    keywords="i2c driver python hardware diy iot raspberry pi",
)
